# ActivitySim
# See full license in LICENSE.txt.

import logging

import openmatrix as omx
import pandas as pd
import numpy as np

from activitysim.core import config
from activitysim.core import inject
from activitysim.core import pipeline

from .util import expressions
from .util.expressions import skim_time_period_label

logger = logging.getLogger(__name__)


@inject.step()
def write_trip_matrices(trips, skim_dict, skim_stack):
    """
    Write trip matrices step.

    Adds boolean columns to local trips table via annotation expressions,
    then aggregates trip counts and writes OD matrices to OMX.  Save annotated
    trips table to pipeline if desired.
    """

    model_settings = config.read_model_settings('write_trip_matrices.yaml')
    trips_df = annotate_trips(trips, skim_dict, skim_stack, model_settings)

    if bool(model_settings.get('SAVE_TRIPS_TABLE')):
        pipeline.replace_table('trips', trips_df)

    logger.info('Aggregating trips...')
    aggregate_trips = trips_df.groupby(['origin', 'destination'], sort=False).sum()

    # use the average household weight for all trips in the origin destination pair
    hh_weight_col = model_settings.get('HH_EXPANSION_WEIGHT_COL')
    aggregate_weight = trips_df[['origin', 'destination', hh_weight_col]].groupby(['origin', 'destination'],
                                                                                  sort=False).mean()
    aggregate_trips[hh_weight_col] = aggregate_weight[hh_weight_col]

    logger.info('Finished.')

    orig_vals = aggregate_trips.index.get_level_values('origin')
    dest_vals = aggregate_trips.index.get_level_values('destination')

    zone_index = pipeline.get_table('land_use').index
    assert all(zone in zone_index for zone in orig_vals)
    assert all(zone in zone_index for zone in dest_vals)

    _, orig_index = zone_index.reindex(orig_vals)
    _, dest_index = zone_index.reindex(dest_vals)

    write_matrices(aggregate_trips, zone_index, orig_index, dest_index, model_settings)


def annotate_trips(trips, skim_dict, skim_stack, model_settings):
    """
    Add columns to local trips table. The annotator has
    access to the origin/destination skims and everything
    defined in the model settings CONSTANTS.

    Pipeline tables can also be accessed by listing them under
    TABLES in the preprocessor settings.
    """

    trips_df = trips.to_frame()

    trace_label = 'trip_matrices'

    # setup skim keys
    assert ('trip_period' not in trips_df)
    trips_df['trip_period'] = skim_time_period_label(trips_df.depart)
    od_skim_wrapper = skim_dict.wrap('origin', 'destination')
    odt_skim_stack_wrapper = skim_stack.wrap(left_key='origin', right_key='destination',
                                             skim_key='trip_period')
    skims = {
        'od_skims': od_skim_wrapper,
        "odt_skims": odt_skim_stack_wrapper
    }

    locals_dict = {}
    constants = config.get_model_constants(model_settings)
    if constants is not None:
        locals_dict.update(constants)

    expressions.annotate_preprocessors(
        trips_df, locals_dict, skims,
        model_settings, trace_label)

    # Data will be expanded by an expansion weight column from
    # the households pipeline table, if specified in the model settings.
    hh_weight_col = model_settings.get('HH_EXPANSION_WEIGHT_COL')

    if hh_weight_col and hh_weight_col not in trips_df:
        logger.info("adding '%s' from households to trips table" % hh_weight_col)
        household_weights = pipeline.get_table('households')[hh_weight_col]
        trips_df[hh_weight_col] = trips_df.household_id.map(household_weights)

    return trips_df


def write_matrices(aggregate_trips, zone_index, orig_index, dest_index, model_settings):
    """
    Write aggregated trips to OMX format.

    The MATRICES setting lists the new OMX files to write.
    Each file can contain any number of 'tables', each specified by a
    table key ('name') and a trips table column ('data_field') to use
    for aggregated counts.

    Any data type may be used for columns added in the annotation phase,
    but the table 'data_field's must be summable types: ints, floats, bools.
    """

    matrix_settings = model_settings.get('MATRICES')

    if not matrix_settings:
        logger.error('Missing MATRICES setting in write_trip_matrices.yaml')

    for matrix in matrix_settings:
        filename = matrix.get('file_name')
        filepath = config.output_file_path(filename)
        logger.info('opening %s' % filepath)
        file = omx.open_file(filepath, 'w')  # possibly overwrite existing file
        table_settings = matrix.get('tables')

        for table in table_settings:
            table_name = table.get('name')
            col = table.get('data_field')

            if col not in aggregate_trips:
                logger.error(f'missing {col} column in aggregate_trips DataFrame')
                return

            hh_weight_col = model_settings.get('HH_EXPANSION_WEIGHT_COL')
            if hh_weight_col:
                aggregate_trips[col] = aggregate_trips[col] / aggregate_trips[hh_weight_col]

            data = np.zeros((len(zone_index), len(zone_index)))
            data[orig_index, dest_index] = aggregate_trips[col]
            logger.info('writing %s' % table_name)
            file[table_name] = data  # write to file

        # include the index-to-zone map in the file
        logger.info('adding %s mapping for %s zones to %s' %
                    (zone_index.name, zone_index.size, filename))
        file.create_mapping(zone_index.name, zone_index.to_numpy())

        logger.info('closing %s' % filepath)
        file.close()



# ActivitySim
# See full license in LICENSE.txt.
from builtins import range
from builtins import int

import sys
import os
import logging
import multiprocessing

from collections import OrderedDict
from functools import reduce
from operator import mul

import numpy as np
import openmatrix as omx

from activitysim.core import skim
from activitysim.core import inject
from activitysim.core import util
from activitysim.core import config
from activitysim.core import tracing

logger = logging.getLogger(__name__)

"""
Read in the omx files and create the skim objects
"""


def get_skim_info(omx_file_path, tags_to_load=None):

    # this is sys.maxint for p2.7 but no limit for p3
    # windows sys.maxint =  2147483647
    MAX_BLOCK_BYTES = sys.maxint - 1 if sys.version_info < (3,) else sys.maxsize - 1

    # Note: we load all skims except those with key2 not in tags_to_load
    # Note: we require all skims to be of same dtype so they can share buffer - is that ok?
    # fixme is it ok to require skims be all the same type? if so, is this the right choice?
    skim_dtype = np.float32
    omx_name = os.path.splitext(os.path.basename(omx_file_path))[0]

    with omx.open_file(omx_file_path) as omx_file:
        # omx_shape = tuple(map(int, tuple(omx_file.shape())))  # sometimes omx shape are floats!

        # fixme call to omx_file.shape() failing in windows p3.5
        omx_shape = omx_file.shape()
        omx_shape = (int(omx_shape[0]), int(omx_shape[1]))  # sometimes omx shape are floats!

        omx_skim_names = omx_file.listMatrices()

        offset_map = None
        offset_map_name = None
        for m in omx_file.listMappings():
            if offset_map is None:
                offset_map_name = m
                offset_map = omx_file.mapentries(offset_map_name)
                assert len(offset_map) == omx_shape[0]

                logger.debug(f"get_skim_info omx_name {omx_name} using offset_map {m}")
            else:
                # don't really expect more than one, but ok if they are all the same
                assert((offset_map == omx_file.mapentries(m).all())), "Multiple different mappings in omx file"

    # - omx_keys dict maps skim key to omx_key
    # DISTWALK: DISTWALK
    # ('DRV_COM_WLK_BOARDS', 'AM'): DRV_COM_WLK_BOARDS__AM, ...
    omx_keys = OrderedDict()
    for skim_name in omx_skim_names:
        key1, sep, key2 = skim_name.partition('__')

        # - ignore composite tags not in tags_to_load
        if tags_to_load and sep and key2 not in tags_to_load:
            continue

        skim_key = (key1, key2) if sep else key1
        omx_keys[skim_key] = skim_name

    num_skims = len(omx_keys)

    # - key1_subkeys dict maps key1 to dict of subkeys with that key1
    # DIST: {'DIST': 0}
    # DRV_COM_WLK_BOARDS: {'MD': 1, 'AM': 0, 'PM': 2}, ...
    key1_subkeys = OrderedDict()
    for skim_key, omx_key in omx_keys.items():
        if isinstance(skim_key, tuple):
            key1, key2 = skim_key
        else:
            key1 = key2 = skim_key
        key2_dict = key1_subkeys.setdefault(key1, {})
        key2_dict[key2] = len(key2_dict)

    # - blocks dict maps block name to blocksize (number of subkey skims in block)
    # skims_0: 198,
    # skims_1: 198, ...
    # - key1_block_offsets dict maps key1 to (block, offset) of first skim with that key1
    # DISTWALK: (0, 2),
    # DRV_COM_WLK_BOARDS: (0, 3), ...

    if MAX_BLOCK_BYTES:
        max_block_items = MAX_BLOCK_BYTES // np.dtype(skim_dtype).itemsize
        max_skims_per_block = max_block_items // multiply_large_numbers(omx_shape)
    else:
        max_skims_per_block = num_skims

    def block_name(block):
        return "skim_%s_%s" % (omx_name, block)

    key1_block_offsets = OrderedDict()
    blocks = OrderedDict()
    block = offset = 0
    for key1, v in key1_subkeys.items():
        num_subkeys = len(v)
        if offset + num_subkeys > max_skims_per_block:  # next block
            blocks[block_name(block)] = offset
            block += 1
            offset = 0
        key1_block_offsets[key1] = (block, offset)
        offset += num_subkeys
    blocks[block_name(block)] = offset  # last block

    # - block_offsets dict maps skim_key to (block, offset) of omx matrix
    # DIST: (0, 0),
    # ('DRV_COM_WLK_BOARDS', 'AM'): (0, 3),
    # ('DRV_COM_WLK_BOARDS', 'MD') (0, 4), ...
    block_offsets = OrderedDict()
    for skim_key in omx_keys:

        if isinstance(skim_key, tuple):
            key1, key2 = skim_key
        else:
            key1 = key2 = skim_key

        block, key1_offset = key1_block_offsets[key1]

        key2_relative_offset = key1_subkeys.get(key1).get(key2)

        block_offsets[skim_key] = (block, key1_offset + key2_relative_offset)

    logger.debug("get_skim_info from %s" % (omx_file_path, ))
    logger.debug("get_skim_info skim_dtype %s omx_shape %s num_skims %s num_blocks %s" %
                 (skim_dtype, omx_shape, num_skims, len(blocks)))

    skim_info = {
        'omx_name': omx_name,
        'omx_shape': omx_shape,
        'num_skims': num_skims,
        'dtype': skim_dtype,
        'offset_map_name': offset_map_name,
        'offset_map': offset_map,
        'omx_keys': omx_keys,
        'key1_block_offsets': key1_block_offsets,
        'block_offsets': block_offsets,
        'blocks': blocks,
    }

    return skim_info


def buffers_for_skims(skim_info, shared=False):

    skim_dtype = skim_info['dtype']
    omx_shape = skim_info['omx_shape']
    blocks = skim_info['blocks']

    skim_buffers = {}
    for block_name, block_size in blocks.items():

        # buffer_size must be int, not np.int64
        buffer_size = int(multiply_large_numbers(omx_shape) * block_size)

        itemsize = np.dtype(skim_dtype).itemsize
        csz = buffer_size * itemsize
        logger.info("allocating shared buffer %s for %s skims (skim size: %s * %s bytes = %s) total size: %s (%s)" %
                    (block_name, block_size, omx_shape, itemsize, buffer_size, csz, util.GB(csz)))

        if shared:
            if np.issubdtype(skim_dtype, np.float64):
                typecode = 'd'
            elif np.issubdtype(skim_dtype, np.float32):
                typecode = 'f'
            else:
                raise RuntimeError("buffers_for_skims unrecognized dtype %s" % skim_dtype)

            buffer = multiprocessing.RawArray(typecode, buffer_size)
        else:
            buffer = np.zeros(buffer_size, dtype=skim_dtype)

        skim_buffers[block_name] = buffer

    return skim_buffers


def skim_data_from_buffers(skim_buffers, skim_info):

    assert type(skim_buffers) == dict

    omx_shape = skim_info['omx_shape']
    skim_dtype = skim_info['dtype']
    blocks = skim_info['blocks']

    skim_data = []
    for block_name, block_size in blocks.items():
        skims_shape = omx_shape + (block_size,)
        block_buffer = skim_buffers[block_name]
        assert len(block_buffer) == int(multiply_large_numbers(skims_shape))
        block_data = np.frombuffer(block_buffer, dtype=skim_dtype).reshape(skims_shape)
        skim_data.append(block_data)

    return skim_data


def default_skim_cache_dir():
    return inject.get_injectable('output_dir')


def build_skim_cache_file_name(omx_name, block):
    return f"cached_{omx_name}_{block}.mmap"


def read_skim_cache(skim_info, skim_data):
    """
        read cached memmapped skim data from canonically named cache file(s) in output directory into skim_data
    """

    skim_cache_dir = config.setting('skim_cache_dir', default_skim_cache_dir())
    logger.info(f"load_skims reading skims data from cache directory {skim_cache_dir}")

    omx_name = skim_info['omx_name']
    dtype = np.dtype(skim_info['dtype'])

    blocks = skim_info['blocks']
    block = 0
    for block_name, block_size in blocks.items():
        skim_cache_file_name = build_skim_cache_file_name(omx_name, block)
        skim_cache_path = os.path.join(skim_cache_dir, skim_cache_file_name)

        assert os.path.isfile(skim_cache_path), \
            "read_skim_cache could not find skim_cache_path: %s" % (skim_cache_path, )

        block_data = skim_data[block]

        logger.info(f"load_skims reading block_name {block_name} {block_data.shape} from {skim_cache_file_name}")

        data = np.memmap(skim_cache_path, shape=block_data.shape, dtype=dtype, mode='r')
        assert data.shape == block_data.shape

        block_data[::] = data[::]

        block += 1


def write_skim_cache(skim_info, skim_data):
    """
        write skim data from skim_data to canonically named cache file(s) in output directory
    """

    skim_cache_dir = config.setting('skim_cache_dir', default_skim_cache_dir())
    logger.info(f"load_skims writing skims data to cache directory {skim_cache_dir}")

    omx_name = skim_info['omx_name']
    dtype = np.dtype(skim_info['dtype'])

    blocks = skim_info['blocks']
    block = 0
    for block_name, block_size in blocks.items():
        skim_cache_file_name = build_skim_cache_file_name(omx_name, block)
        skim_cache_path = os.path.join(skim_cache_dir, skim_cache_file_name)

        block_data = skim_data[block]

        logger.info(f"load_skims writing block_name {block_name} {block_data.shape} to {skim_cache_file_name}")

        data = np.memmap(skim_cache_path, shape=block_data.shape, dtype=dtype, mode='w+')
        data[::] = block_data

        block += 1


def read_skims_from_omx(skim_info, skim_data, omx_file_path):
    """
    read skims from omx file into skim_data
    """

    block_offsets = skim_info['block_offsets']
    omx_keys = skim_info['omx_keys']

    # read skims into skim_data
    with omx.open_file(omx_file_path) as omx_file:
        for skim_key, omx_key in omx_keys.items():

            omx_data = omx_file[omx_key]
            assert np.issubdtype(omx_data.dtype, np.floating)

            block, offset = block_offsets[skim_key]
            block_data = skim_data[block]

            logger.debug("load_skims load omx_key %s skim_key %s to block %s offset %s" %
                         (omx_key, skim_key, block, offset))

            # this will trigger omx readslice to read and copy data to skim_data's buffer
            a = block_data[:, :, offset]
            a[:] = omx_data[:]

    logger.info("load_skims loaded skims from %s" % (omx_file_path, ))


def load_skims(omx_file_path, skim_info, skim_buffers):

    read_cache = config.setting('read_skim_cache')
    write_cache = config.setting('write_skim_cache')
    assert not (read_cache and write_cache), \
        "read_skim_cache and write_skim_cache are both True in settings file. I am assuming this is a mistake"

    skim_data = skim_data_from_buffers(skim_buffers, skim_info)

    t0 = tracing.print_elapsed_time()

    if read_cache:
        read_skim_cache(skim_info, skim_data)
        t0 = tracing.print_elapsed_time("read_skim_cache", t0)
    else:
        read_skims_from_omx(skim_info, skim_data, omx_file_path)
        t0 = tracing.print_elapsed_time("read_skims_from_omx", t0)

    if write_cache:
        write_skim_cache(skim_info, skim_data)
        t0 = tracing.print_elapsed_time("write_skim_cache", t0)


@inject.injectable(cache=True)
def skim_dict(settings):

    omx_file_path = config.data_file_path(settings["skims_file"])
    tags_to_load = settings['skim_time_periods']['labels']

    logger.info("loading skim_dict from %s" % (omx_file_path, ))

    # select the skims to load
    skim_info = get_skim_info(omx_file_path, tags_to_load)

    logger.debug("omx_shape %s skim_dtype %s" % (skim_info['omx_shape'], skim_info['dtype']))

    skim_buffers = inject.get_injectable('data_buffers', None)
    if skim_buffers:
        logger.info('Using existing skim_buffers for skims')
    else:
        skim_buffers = buffers_for_skims(skim_info, shared=False)
        load_skims(omx_file_path, skim_info, skim_buffers)

    skim_data = skim_data_from_buffers(skim_buffers, skim_info)

    block_names = list(skim_info['blocks'].keys())
    for i in range(len(skim_data)):
        block_name = block_names[i]
        block_data = skim_data[i]
        logger.info("block_name %s bytes %s (%s)" %
                    (block_name, block_data.nbytes, util.GB(block_data.nbytes)))

    # create skim dict
    skim_dict = skim.SkimDict(skim_data, skim_info)

    offset_map = skim_info['offset_map']
    if offset_map is not None:
        skim_dict.offset_mapper.set_offset_list(offset_map)
        logger.debug(f"using offset map {skim_info['offset_map_name']}from omx file: {offset_map}")
    else:
        # assume this is a one-based skim map
        skim_dict.offset_mapper.set_offset_int(-1)

    return skim_dict


def multiply_large_numbers(list_of_numbers):
    return reduce(mul, list_of_numbers)


@inject.injectable(cache=True)
def skim_stack(skim_dict):

    logger.debug("loading skim_stack injectable")
    return skim.SkimStack(skim_dict)



# ActivitySim
# See full license in LICENSE.txt.
from builtins import input

import logging
from collections import OrderedDict

import numpy as np
import pandas as pd

from . import util
from . import mem

logger = logging.getLogger(__name__)

# dict of table_dicts keyed by trace_label
# table_dicts are dicts tuples of {table_name: (elements, bytes, mem), ...}
CHUNK_LOG = OrderedDict()

# array of chunk_size active CHUNK_LOG
CHUNK_SIZE = []
EFFECTIVE_CHUNK_SIZE = []

HWM = [{}]


def GB(bytes):
    # symbols = ('', 'K', 'M', 'G', 'T')
    symbols = ('', ' KB', ' MB', ' GB', ' TB')
    fmt = "%.1f%s"
    for i, s in enumerate(symbols):
        units = 1 << i * 10
        if bytes < units * 1024:
            return fmt % (bytes / units, s)


def commas(x):
    x = int(x)
    if x < 10000:
        return str(x)
    result = ''
    while x >= 1000:
        x, r = divmod(x, 1000)
        result = ",%03d%s" % (r, result)
    return "%d%s" % (x, result)


def log_open(trace_label, chunk_size, effective_chunk_size):

    # nested chunkers should be unchunked
    if len(CHUNK_LOG) > 0:
        assert chunk_size == 0
        assert trace_label not in CHUNK_LOG

    logger.debug("log_open chunker %s chunk_size %s effective_chunk_size %s" %
                 (trace_label, commas(chunk_size), commas(effective_chunk_size)))

    CHUNK_LOG[trace_label] = OrderedDict()
    CHUNK_SIZE.append(chunk_size)
    EFFECTIVE_CHUNK_SIZE.append(effective_chunk_size)

    HWM.append({})


def log_close(trace_label):

    assert CHUNK_LOG and next(reversed(CHUNK_LOG)) == trace_label

    logger.debug("log_close %s" % trace_label)

    # if we are closing base level chunker
    if len(CHUNK_LOG) == 1:
        log_write_hwm()

    label, _ = CHUNK_LOG.popitem(last=True)
    assert label == trace_label
    CHUNK_SIZE.pop()
    EFFECTIVE_CHUNK_SIZE.pop()

    HWM.pop()


def log_df(trace_label, table_name, df):

    if df is None:
        # FIXME force_garbage_collect on delete?
        mem.force_garbage_collect()

    cur_chunker = next(reversed(CHUNK_LOG))

    if df is None:
        CHUNK_LOG.get(cur_chunker).pop(table_name)
        op = 'del'

        logger.debug("log_df del %s : %s " % (table_name, trace_label))

    else:

        shape = df.shape
        elements = np.prod(shape, dtype=np.int64)
        op = 'add'

        if isinstance(df, pd.Series):
            bytes = df.memory_usage(index=True)
        elif isinstance(df, pd.DataFrame):
            bytes = df.memory_usage(index=True).sum()
        elif isinstance(df, np.ndarray):
            bytes = df.nbytes
        else:
            logger.error("log_df %s unknown type: %s" % (table_name, type(df)))
            assert False

        CHUNK_LOG.get(cur_chunker)[table_name] = (elements, bytes)

        # log this df
        logger.debug("log_df add %s elements: %s bytes: %s shape: %s : %s " %
                     (table_name, commas(elements), GB(bytes), shape, trace_label))

    total_elements, total_bytes = _chunk_totals()  # new chunk totals
    cur_mem = mem.get_memory_info()
    hwm_trace_label = "%s.%s.%s" % (trace_label, op, table_name)

    logger.debug("total_elements: %s, total_bytes: %s cur_mem: %s: %s " %
                 (total_elements, GB(total_bytes), GB(cur_mem), hwm_trace_label))

    mem.trace_memory_info(hwm_trace_label)

    # - check high_water_marks

    info = "elements: %s bytes: %s mem: %s chunk_size: %s effective_chunk_size: %s" % \
           (commas(total_elements), GB(total_bytes), GB(cur_mem),
            commas(CHUNK_SIZE[0]), commas(EFFECTIVE_CHUNK_SIZE[0]))

    check_hwm('elements', total_elements, info, hwm_trace_label)
    check_hwm('bytes', total_bytes, info, hwm_trace_label)
    check_hwm('mem', cur_mem, info, hwm_trace_label)


def _chunk_totals():

    total_elements = 0
    total_bytes = 0
    for label in CHUNK_LOG:
        tables = CHUNK_LOG[label]
        for table_name in tables:
            elements, bytes = tables[table_name]
            total_elements += elements
            total_bytes += bytes

    return total_elements, total_bytes


def check_hwm(tag, value, info, trace_label):

    for d in HWM:

        hwm = d.setdefault(tag, {})

        if value > hwm.get('mark', 0):
            hwm['mark'] = value
            hwm['info'] = info
            hwm['trace_label'] = trace_label


def log_write_hwm():

    d = HWM[0]
    for tag in d:
        hwm = d[tag]
        logger.debug("#chunk_hwm high_water_mark %s: %s (%s) in %s" %
                     (tag, hwm['mark'], hwm['info'], hwm['trace_label']), )

    # - elements shouldn't exceed chunk_size or effective_chunk_size of base chunker
    def check_chunk_size(hwm, chunk_size, label, max_leeway):
        elements = hwm['mark']
        if chunk_size and max_leeway and elements > chunk_size * max_leeway:  # too high
            # FIXME check for #warning in log - there is nothing the user can do about this
            logger.debug("#chunk_hwm #warning total_elements (%s) > %s (%s) %s : %s " %
                         (commas(elements), label, commas(chunk_size),
                          hwm['info'], hwm['trace_label']))

    # if we are in a chunker
    if len(HWM) > 1 and HWM[1]:
        assert 'elements' in HWM[1]  # expect an 'elements' hwm dict for base chunker
        hwm = HWM[1].get('elements')
        check_chunk_size(hwm, EFFECTIVE_CHUNK_SIZE[0],  'effective_chunk_size', max_leeway=1.1)
        check_chunk_size(hwm, CHUNK_SIZE[0], 'chunk_size', max_leeway=1)


def rows_per_chunk(chunk_size, row_size, num_choosers, trace_label):

    if chunk_size > 0:
        # closest number of chooser rows to achieve chunk_size without exceeding
        max_rpc = int(chunk_size / float(row_size))
    else:
        max_rpc = num_choosers

    rpc = int(np.clip(max_rpc, 1, num_choosers))

    # chunks = int(ceil(num_choosers / float(rpc)))
    effective_chunk_size = row_size * rpc
    num_chunks = (num_choosers // rpc) + (num_choosers % rpc > 0)

    logger.debug(f"#chunk_calc num_chunks: {num_chunks}, rows_per_chunk: {rpc}, "
                 f"max_rpc: {max_rpc}, "
                 f"effective_chunk_size: {effective_chunk_size}, num_choosers: {num_choosers} : {trace_label}")

    return rpc, effective_chunk_size


def chunked_choosers(choosers, rows_per_chunk):

    assert choosers.shape[0] > 0

    # generator to iterate over choosers in chunk_size chunks
    num_choosers = len(choosers.index)
    num_chunks = (num_choosers // rows_per_chunk) + (num_choosers % rows_per_chunk > 0)

    i = offset = 0
    while offset < num_choosers:
        yield i+1, num_chunks, choosers.iloc[offset: offset+rows_per_chunk]
        offset += rows_per_chunk
        i += 1


def chunked_choosers_and_alts(choosers, alternatives, rows_per_chunk):
    """
    generator to iterate over choosers and alternatives in chunk_size chunks

    like chunked_choosers, but also chunks alternatives
    for use with sampled alternatives which will have different alternatives (and numbers of alts)

    There may be up to sample_size (or as few as one) alternatives for each chooser
    because alternatives may have been sampled more than once,  but pick_count for those
    alternatives will always sum to sample_size.

    When we chunk the choosers, we need to take care chunking the alternatives as there are
    varying numbers of them for each chooser. Since alternatives appear in the same order
    as choosers, we can use cumulative pick_counts to identify boundaries of sets of alternatives

    Parameters
    ----------
    choosers
    alternatives : pandas DataFrame
        sample alternatives including pick_count column in same order as choosers
    rows_per_chunk : int

    Yields
    -------
    i : int
        one-based index of current chunk
    num_chunks : int
        total number of chunks that will be yielded
    choosers : pandas DataFrame slice
        chunk of choosers
    alternatives : pandas DataFrame slice
        chunk of alternatives for chooser chunk
    """

    # if not choosers.index.is_monotonic_increasing:
    #     logger.warning('sorting choosers because not monotonic increasing')
    #     choosers = choosers.sort_index()

    # alternatives index should match choosers (except with duplicate repeating alt rows)
    assert choosers.index.equals(alternatives.index[~alternatives.index.duplicated(keep='first')])

    last_repeat = alternatives.index != np.roll(alternatives.index, -1)
    assert (choosers.shape[0] == 1) or choosers.index.equals(alternatives.index[last_repeat])

    assert choosers.shape[0] > 0
    assert 'pick_count' in alternatives.columns or choosers.index.name == alternatives.index.name

    num_choosers = len(choosers.index)
    num_chunks = (num_choosers // rows_per_chunk) + (num_choosers % rows_per_chunk > 0)

    assert choosers.index.name == alternatives.index.name

    # alt chunks boundaries are where index changes
    alt_ids = alternatives.index.values
    alt_chunk_end = np.where(alt_ids[:-1] != alt_ids[1:])[0] + 1
    alt_chunk_end = np.append([0], alt_chunk_end)  # including the first...
    alt_chunk_end = alt_chunk_end[rows_per_chunk::rows_per_chunk]

    # add index to end of array to capture any final partial chunk
    alt_chunk_end = np.append(alt_chunk_end, [len(alternatives.index)])

    i = offset = alt_offset = 0
    while offset < num_choosers:

        alt_end = alt_chunk_end[i]

        chooser_chunk = choosers[offset: offset + rows_per_chunk]
        alternative_chunk = alternatives[alt_offset: alt_end]

        assert len(chooser_chunk.index) == len(np.unique(alternative_chunk.index.values))

        yield i+1, num_chunks, chooser_chunk, alternative_chunk

        i += 1
        offset += rows_per_chunk
        alt_offset = alt_end


def chunked_choosers_by_chunk_id(choosers, rows_per_chunk):
    # generator to iterate over choosers in chunk_size chunks
    # like chunked_choosers but based on chunk_id field rather than dataframe length
    # (the presumption is that choosers has multiple rows with the same chunk_id that
    # all have to be included in the same chunk)
    # FIXME - we pathologically know name of chunk_id col in households table

    assert choosers.shape[0] > 0

    num_choosers = choosers['chunk_id'].max() + 1
    num_chunks = (num_choosers // rows_per_chunk) + (num_choosers % rows_per_chunk > 0)

    i = offset = 0
    while offset < num_choosers:
        chooser_chunk = choosers[choosers['chunk_id'].between(offset, offset + rows_per_chunk - 1)]
        yield i+1, num_chunks, chooser_chunk
        offset += rows_per_chunk
        i += 1



from collections import OrderedDict


import numpy as np
import pytest

from activitysim.abm.tables import skims


@pytest.fixture(scope="session")
def matrix_dimension():
    return 5922


@pytest.fixture(scope="session")
def num_of_matrices():
    return 845


@pytest.fixture(scope="session")
def skim_info(num_of_matrices, matrix_dimension):
    time_periods = ['EA', 'AM', 'MD', 'PM', 'NT']

    omx_keys = OrderedDict()
    omx_key1_block_offsets = OrderedDict()
    omx_block_offsets = OrderedDict()
    omx_blocks = OrderedDict()
    omx_blocks['skim_arc_skims_0'] = num_of_matrices

    for i in range(0, num_of_matrices + 1):
        key1_name = 'm{}'.format(i // len(time_periods) + 1)
        time_period = time_periods[i % len(time_periods)]

        omx_keys[(key1_name, time_period)] = '{}__{}'.format(key1_name, time_period)
        omx_block_offsets[(key1_name, time_period)] = (0, i)

        if 0 == i % len(time_periods):
            omx_key1_block_offsets[key1_name] = (0, i)

    skim_info = {
        'omx_name': 'arc_skims',
        'omx_shape': (matrix_dimension, matrix_dimension),
        'num_skims': num_of_matrices,
        'dtype': np.float32,
        'omx_keys': omx_keys,
        'key1_block_offsets': omx_key1_block_offsets,
        'block_offsets': omx_block_offsets,
        'blocks': omx_blocks
    }

    return skim_info


def test_multiply_large_numbers(skim_info, num_of_matrices, matrix_dimension):
    omx_shape = skim_info['omx_shape']
    blocks = skim_info['blocks']

    for block_name, block_size in blocks.items():
        # If overflow, this number will go negative
        assert int(skims.multiply_large_numbers(omx_shape) * block_size) == \
               num_of_matrices * matrix_dimension ** 2


def test_multiple_large_floats():
    calculated_value = skims.multiply_large_numbers([6205.1, 5423.2, 932.4, 15.4])
    actual_value = 483200518316.9472
    assert abs(calculated_value - actual_value) < 0.0001



# ActivitySim
# See full license in LICENSE.txt.
import logging
import sys
import pandas as pd

from collections import OrderedDict

from activitysim.core import pipeline
from activitysim.core import inject
from activitysim.core import config

from activitysim.core.config import setting

logger = logging.getLogger(__name__)


def track_skim_usage(output_dir):
    """
    write statistics on skim usage (diagnostic to detect loading of un-needed skims)

    FIXME - have not yet implemented a facility to avoid loading of unused skims

    Parameters
    ----------
    output_dir: str

    """
    pd.options.display.max_columns = 500
    pd.options.display.max_rows = 100

    skim_dict = inject.get_injectable('skim_dict')
    skim_stack = inject.get_injectable('skim_stack', None)

    mode = 'wb' if sys.version_info < (3,) else 'w'
    with open(config.output_file_path('skim_usage.txt'), mode) as output_file:

        print("\n### skim_dict usage", file=output_file)
        for key in skim_dict.usage:
            print(key, file=output_file)

        if skim_stack is None:

            unused_keys = {k for k in skim_dict.skim_info['omx_keys']} - \
                          {k for k in skim_dict.usage}

            print("\n### unused skim keys", file=output_file)
            for key in unused_keys:
                print(key, file=output_file)

        else:

            print("\n### skim_stack usage", file=output_file)
            for key in skim_stack.usage:
                print(key, file=output_file)

            unused = {k for k in skim_dict.skim_info['omx_keys'] if not isinstance(k, tuple)} - \
                     {k for k in skim_dict.usage if not isinstance(k, tuple)}
            print("\n### unused skim str keys", file=output_file)
            for key in unused:
                print(key, file=output_file)

                unused = {k[0] for k in skim_dict.skim_info['omx_keys'] if isinstance(k, tuple)} - \
                         {k[0] for k in skim_dict.usage if isinstance(k, tuple)} - \
                         {k for k in skim_stack.usage}
            print("\n### unused skim dim3 keys", file=output_file)
            for key in unused:
                print(key, file=output_file)


def write_data_dictionary(output_dir):
    """
    Write table_name, number of rows, columns, and bytes for each checkpointed table

    Parameters
    ----------
    output_dir: str

    """
    pd.options.display.max_columns = 500
    pd.options.display.max_rows = 100

    output_tables = pipeline.checkpointed_tables()

    # write data dictionary for all checkpointed_tables

    mode = 'wb' if sys.version_info < (3,) else 'w'
    with open(config.output_file_path('data_dict.txt'), mode) as output_file:
        for table_name in output_tables:
            df = inject.get_table(table_name, None).to_frame()

            print("\n### %s %s" % (table_name, df.shape), file=output_file)
            print('index:', df.index.name, df.index.dtype, file=output_file)
            print(df.dtypes, file=output_file)


# def xwrite_data_dictionary(output_dir):
#     """
#     Write table_name, number of rows, columns, and bytes for each checkpointed table
#
#     Parameters
#     ----------
#     output_dir: str
#
#     """
#     pd.options.display.max_columns = 500
#     pd.options.display.max_rows = 100
#
#     checkpoints = pipeline.get_checkpoints()
#     tables = OrderedDict()
#
#     table_names = [c for c in checkpoints if c not in pipeline.NON_TABLE_COLUMNS]
#
#     with open(config.output_file_path('data_dict.txt'), 'wb') as file:
#
#         for index, row in checkpoints.iterrows():
#
#             checkpoint = row[pipeline.CHECKPOINT_NAME]
#
#             print("\n##########################################", file=file)
#             print("# %s" % checkpoint, file=file)
#             print("##########################################", file=file)
#
#             for table_name in table_names:
#
#                 if row[table_name] == '' and table_name in tables:
#                     print("\n### %s dropped %s" % (checkpoint, table_name, ), file=file)
#                     del tables[table_name]
#
#                 if row[table_name] == checkpoint:
#                     df = pipeline.get_table(table_name, checkpoint)
#                     info = tables.get(table_name, None)
#                     if info is None:
#
#                         print("\n### %s created %s %s\n" %
#                               (checkpoint, table_name, df.shape), file=file)
#
#                         print(df.dtypes, file=file)
#                         print('index:', df.index.name, df.index.dtype, file=file)
#
#                     else:
#                         new_cols = [c for c in df.columns.values if c not in info['columns']]
#                         dropped_cols = [c for c in info['columns'] if c not in df.columns.values]
#                         new_rows = df.shape[0] - info['num_rows']
#                         if new_cols:
#
#                             print("\n### %s added %s columns to %s %s\n" %
#                                   (checkpoint, len(new_cols), table_name, df.shape), file=file)
#                             print(df[new_cols].dtypes, file=file)
#
#                         if dropped_cols:
#                             print("\n### %s dropped %s columns from %s %s\n" %
#                                   (checkpoint,  len(dropped_cols), table_name, df.shape),
#                                   file=file)
#                             print(dropped_cols, file=file)
#
#                         if new_rows > 0:
#                             print("\n### %s added %s rows to %s %s" %
#                                   (checkpoint, new_rows, table_name, df.shape), file=file)
#                         elif new_rows < 0:
#                             print("\n### %s dropped %s rows from %s %s" %
#                                   (checkpoint, new_rows, table_name, df.shape), file=file)
#                         else:
#                             if not new_cols and not dropped_cols:
#                                 print("\n### %s modified %s %s" %
#                                       (checkpoint, table_name, df.shape), file=file)
#
#                     tables[table_name] = {
#                         'checkpoint_name': checkpoint,
#                         'columns': df.columns.values,
#                         'num_rows': df.shape[0]
#                     }


def write_tables(output_dir):
    """
    Write pipeline tables as csv files (in output directory) as specified by output_tables list
    in settings file.

    'output_tables' can specify either a list of output tables to include or to skip
    if no output_tables list is specified, then no checkpointed tables will be written

    To write all output tables EXCEPT the households and persons tables:

    ::

      output_tables:
        action: skip
        tables:
          - households
          - persons

    To write ONLY the households table:

    ::

      output_tables:
        action: include
        tables:
           - households

    To write tables into a single HDF5 store instead of individual CSVs, use the h5_store flag:

    ::

      output_tables:
        h5_store: True
        action: include
        tables:
           - households

    Parameters
    ----------
    output_dir: str

    """

    output_tables_settings_name = 'output_tables'

    output_tables_settings = setting(output_tables_settings_name)

    if output_tables_settings is None:
        logger.info("No output_tables specified in settings file. Nothing to write.")
        return

    action = output_tables_settings.get('action')
    tables = output_tables_settings.get('tables')
    prefix = output_tables_settings.get('prefix', 'final_')
    h5_store = output_tables_settings.get('h5_store', False)

    if action not in ['include', 'skip']:
        raise "expected %s action '%s' to be either 'include' or 'skip'" % \
              (output_tables_settings_name, action)

    checkpointed_tables = pipeline.checkpointed_tables()
    if action == 'include':
        output_tables_list = tables
    elif action == 'skip':
        output_tables_list = [t for t in checkpointed_tables if t not in tables]

    for table_name in output_tables_list:

        if table_name == 'checkpoints':
            df = pipeline.get_checkpoints()
        else:
            if table_name not in checkpointed_tables:
                logger.warning("Skipping '%s': Table not found." % table_name)
                continue
            df = pipeline.get_table(table_name)

        if h5_store:
            file_path = config.output_file_path('%soutput_tables.h5' % prefix)
            df.to_hdf(file_path, key=table_name, mode='a', format='fixed')
        else:
            file_name = "%s%s.csv" % (prefix, table_name)
            file_path = config.output_file_path(file_name)

            # include the index if it has a name or is a MultiIndex
            write_index = df.index.name is not None or isinstance(df.index, pd.MultiIndex)

            df.to_csv(file_path, index=write_index)



# ActivitySim
# See full license in LICENSE.txt.
from builtins import zip
from builtins import range

import logging

import pandas as pd

from activitysim.core import simulate
from activitysim.core import tracing
from activitysim.core import config
from activitysim.core import inject
from activitysim.core import pipeline
from activitysim.core.mem import force_garbage_collect

from .util.expressions import annotate_preprocessors

from activitysim.core import assign
from activitysim.core.util import assign_in_place

from .util.expressions import skim_time_period_label

from .util.mode import mode_choice_simulate

logger = logging.getLogger(__name__)


@inject.step()
def trip_mode_choice(
        trips,
        tours_merged,
        skim_dict, skim_stack,
        chunk_size, trace_hh_id):
    """
    Trip mode choice - compute trip_mode (same values as for tour_mode) for each trip.

    Modes for each primary tour putpose are calculated separately because they have different
    coefficient values (stored in trip_mode_choice_coeffs.csv coefficient file.)

    Adds trip_mode column to trip table
    """
    trace_label = 'trip_mode_choice'
    model_settings = config.read_model_settings('trip_mode_choice.yaml')

    logsum_column_name = model_settings.get('MODE_CHOICE_LOGSUM_COLUMN_NAME')
    mode_column_name = 'trip_mode'

    model_spec = \
        simulate.read_model_spec(file_name=model_settings['SPEC'])
    omnibus_coefficients = \
        assign.read_constant_spec(config.config_file_path(model_settings['COEFFICIENTS']))

    trips_df = trips.to_frame()
    logger.info("Running %s with %d trips", trace_label, trips_df.shape[0])

    tours_merged = tours_merged.to_frame()
    tours_merged = tours_merged[model_settings['TOURS_MERGED_CHOOSER_COLUMNS']]

    nest_spec = config.get_logit_model_settings(model_settings)

    tracing.print_summary('primary_purpose',
                          trips_df.primary_purpose, value_counts=True)

    # - trips_merged - merge trips and tours_merged
    trips_merged = pd.merge(
        trips_df,
        tours_merged,
        left_on='tour_id',
        right_index=True,
        how="left")
    assert trips_merged.index.equals(trips.index)

    # setup skim keys
    assert ('trip_period' not in trips_merged)
    trips_merged['trip_period'] = skim_time_period_label(trips_merged.depart)

    orig_col = 'origin'
    dest_col = 'destination'

    odt_skim_stack_wrapper = skim_stack.wrap(left_key=orig_col, right_key=dest_col,
                                             skim_key='trip_period')
    dot_skim_stack_wrapper = skim_stack.wrap(left_key=dest_col, right_key=orig_col,
                                             skim_key='trip_period')
    od_skim_wrapper = skim_dict.wrap('origin', 'destination')

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "od_skims": od_skim_wrapper,
    }

    constants = config.get_model_constants(model_settings)
    constants.update({
        'ORIGIN': orig_col,
        'DESTINATION': dest_col
    })

    choices_list = []
    for primary_purpose, trips_segment in trips_merged.groupby('primary_purpose'):

        segment_trace_label = tracing.extend_trace_label(trace_label, primary_purpose)

        logger.info("trip_mode_choice tour_type '%s' (%s trips)" %
                    (primary_purpose, len(trips_segment.index), ))

        # name index so tracing knows how to slice
        assert trips_segment.index.name == 'trip_id'

        locals_dict = assign.evaluate_constants(omnibus_coefficients[primary_purpose],
                                                constants=constants)
        locals_dict.update(constants)

        annotate_preprocessors(
            trips_segment, locals_dict, skims,
            model_settings, segment_trace_label)

        locals_dict.update(skims)

        choices = mode_choice_simulate(
            choosers=trips_segment,
            spec=model_spec,
            nest_spec=nest_spec,
            skims=skims,
            locals_d=locals_dict,
            chunk_size=chunk_size,
            mode_column_name=mode_column_name,
            logsum_column_name=logsum_column_name,
            trace_label=trace_label,
            trace_choice_name='trip_mode_choice')

        if trace_hh_id:
            # trace the coefficients
            tracing.trace_df(pd.Series(locals_dict),
                             label=tracing.extend_trace_label(segment_trace_label, 'constants'),
                             transpose=False,
                             slicer='NONE')

            # so we can trace with annotations
            assign_in_place(trips_segment, choices)

            tracing.trace_df(trips_segment,
                             label=tracing.extend_trace_label(segment_trace_label, 'trip_mode'),
                             slicer='tour_id',
                             index_label='tour_id',
                             warn_if_empty=True)

        choices_list.append(choices)

        # FIXME - force garbage collection
        force_garbage_collect()

    choices = pd.concat(choices_list)

    # keep mode_choice and (optionally) logsum columns
    trips_df = trips.to_frame()
    assign_in_place(trips_df, choices)

    tracing.print_summary('tour_modes',
                          trips_merged.tour_mode, value_counts=True)

    tracing.print_summary('trip_mode_choice choices',
                          trips_df[mode_column_name], value_counts=True)

    assert not trips_df[mode_column_name].isnull().any()

    pipeline.replace_table("trips", trips_df)

    if trace_hh_id:
        tracing.trace_df(trips_df,
                         label=tracing.extend_trace_label(trace_label, 'trip_mode'),
                         slicer='trip_id',
                         index_label='trip_id',
                         warn_if_empty=True)



# ActivitySim
# See full license in LICENSE.txt.
import logging

from activitysim.core import simulate
from activitysim.core import tracing
from activitysim.core import config

from activitysim.core.assign import evaluate_constants


from . import expressions


logger = logging.getLogger(__name__)


def filter_chooser_columns(choosers, logsum_settings, model_settings):

    chooser_columns = logsum_settings.get('LOGSUM_CHOOSER_COLUMNS', [])

    if 'CHOOSER_ORIG_COL_NAME' in model_settings:
        chooser_columns.append(model_settings['CHOOSER_ORIG_COL_NAME'])

    missing_columns = [c for c in chooser_columns if c not in choosers]
    if missing_columns:
        logger.debug("logsum.filter_chooser_columns missing_columns %s" % missing_columns)

    # ignore any columns not appearing in choosers df
    chooser_columns = [c for c in chooser_columns if c in choosers]

    choosers = choosers[chooser_columns]
    return choosers


def compute_logsums(choosers,
                    tour_purpose,
                    logsum_settings, model_settings,
                    skim_dict, skim_stack,
                    chunk_size, trace_label):
    """

    Parameters
    ----------
    choosers
    tour_purpose
    logsum_settings
    model_settings
    skim_dict
    skim_stack
    chunk_size
    trace_hh_id
    trace_label

    Returns
    -------
    logsums: pandas series
        computed logsums with same index as choosers
    """

    trace_label = tracing.extend_trace_label(trace_label, 'compute_logsums')

    # compute_logsums needs to know name of dest column in interaction_sample
    orig_col_name = model_settings['CHOOSER_ORIG_COL_NAME']
    dest_col_name = model_settings['ALT_DEST_COL_NAME']

    # FIXME - are we ok with altering choosers (so caller doesn't have to set these)?
    assert ('in_period' not in choosers) and ('out_period' not in choosers)
    choosers['in_period'] = expressions.skim_time_period_label(model_settings['IN_PERIOD'])
    choosers['out_period'] = expressions.skim_time_period_label(model_settings['OUT_PERIOD'])

    assert ('duration' not in choosers)
    choosers['duration'] = model_settings['IN_PERIOD'] - model_settings['OUT_PERIOD']

    logsum_spec = simulate.read_model_spec(file_name=logsum_settings['SPEC'])
    coefficients = simulate.get_segment_coefficients(logsum_settings, tour_purpose)
    logsum_spec = simulate.eval_coefficients(logsum_spec, coefficients, estimator=None)

    nest_spec = config.get_logit_model_settings(logsum_settings)
    nest_spec = simulate.eval_nest_coefficients(nest_spec, coefficients)

    constants = config.get_model_constants(logsum_settings)

    logger.debug("Running compute_logsums with %d choosers" % choosers.shape[0])

    # setup skim keys
    odt_skim_stack_wrapper = skim_stack.wrap(left_key=orig_col_name, right_key=dest_col_name,
                                             skim_key='out_period')
    dot_skim_stack_wrapper = skim_stack.wrap(left_key=dest_col_name, right_key=orig_col_name,
                                             skim_key='in_period')
    odr_skim_stack_wrapper = skim_stack.wrap(left_key=orig_col_name, right_key=dest_col_name,
                                             skim_key='in_period')
    dor_skim_stack_wrapper = skim_stack.wrap(left_key=dest_col_name, right_key=orig_col_name,
                                             skim_key='out_period')
    od_skim_stack_wrapper = skim_dict.wrap(orig_col_name, dest_col_name)

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "odr_skims": odr_skim_stack_wrapper,
        "dor_skims": dor_skim_stack_wrapper,
        "od_skims": od_skim_stack_wrapper,
        'orig_col_name': orig_col_name,
        'dest_col_name': dest_col_name
    }

    locals_dict = {}
    locals_dict.update(constants)
    locals_dict.update(skims)

    # constrained coefficients can appear in expressions
    locals_dict.update(coefficients)

    # - run preprocessor to annotate choosers
    # allow specification of alternate preprocessor for nontour choosers
    preprocessor = model_settings.get('LOGSUM_PREPROCESSOR', 'preprocessor')
    preprocessor_settings = logsum_settings[preprocessor]

    if preprocessor_settings:

        simulate.set_skim_wrapper_targets(choosers, skims)

        expressions.assign_columns(
            df=choosers,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label)

    logsums = simulate.simple_simulate_logsums(
        choosers,
        logsum_spec,
        nest_spec,
        skims=skims,
        locals_d=locals_dict,
        chunk_size=chunk_size,
        trace_label=trace_label,
        alt_col_name=dest_col_name)

    return logsums



# ActivitySim
# See full license in LICENSE.txt.

from builtins import range
from builtins import object

import logging

from collections import OrderedDict

import numpy as np
import pandas as pd

from activitysim.core.util import quick_loc_series


logger = logging.getLogger(__name__)


class OffsetMapper(object):
    """
    Utility to map skim zone ids to ordinal offsets (e.g. numpy array indices)

    Can map either by a fixed offset (e.g. -1 to map 1-based to 0-based)
    or by an explicit mapping of zone id to offset (slower but more flexible)
    """

    def __init__(self, offset_int=None):
        self.offset_series = None
        self.offset_int = offset_int

    def set_offset_list(self, offset_list):
        """
        Specify the zone ids corresponding to the offsets (ordinal positions)

        set_offset_list([10, 20, 30, 40])
        map([30, 20, 40])
        returns offsets [2, 1, 3]

        Parameters
        ----------
        offset_list : list of int
        """
        assert isinstance(offset_list, list)
        assert self.offset_int is None

        # - for performance, check if this is a simple int-based series
        first_offset = offset_list[0]
        if (offset_list == list(range(first_offset, len(offset_list)+first_offset))):
            offset_int = -1 * first_offset
            # print "set_offset_list substituting offset_int of %s" % offset_int
            self.set_offset_int(offset_int)
            return

        if self.offset_series is None:
            self.offset_series = pd.Series(data=list(range(len(offset_list))), index=offset_list)
        else:
            # make sure it offsets are the same
            assert (offset_list == self.offset_series.index).all()

    def set_offset_int(self, offset_int):
        """
        specify fixed offset (e.g. -1 to map 1-based to 0-based)

        Parameters
        ----------
        offset_int : int
        """

        # should be some kind of integer
        assert int(offset_int) == offset_int
        assert self.offset_series is None

        if self.offset_int is None:
            self.offset_int = int(offset_int)
        else:
            # make sure it is the same
            assert offset_int == self.offset_int

    def map(self, zone_ids):
        """
        map zone_ids to offsets

        Parameters
        ----------
        zone_ids

        Returns
        -------
        offsets : numpy array of int
        """

        NOT_IN_SKIM = -1

        if self.offset_series is not None:
            assert(self.offset_int is None)
            assert isinstance(self.offset_series, pd.Series)
            offsets = np.asanyarray(quick_loc_series(zone_ids, self.offset_series).fillna(NOT_IN_SKIM).astype(int))

        elif self.offset_int:
            assert (self.offset_series is None)
            offsets = zone_ids + self.offset_int
        else:
            offsets = zone_ids

        return offsets


class SkimWrapper(object):
    """
    Container for skim arrays.

    Parameters
    ----------
    data : 2D array
    offset : int, optional
        An optional offset that will be added to origin/destination
        values to turn them into array indices.
        For example, if zone IDs are 1-based, an offset of -1
        would turn them into 0-based array indices.

    """
    def __init__(self, data, offset_mapper=None):

        self.data = data
        self.offset_mapper = offset_mapper if offset_mapper is not None else OffsetMapper()

    def get(self, orig, dest):
        """
        Get impedence values for a set of origin, destination pairs.

        Parameters
        ----------
        orig : 1D array
        dest : 1D array

        Returns
        -------
        values : 1D array

        """

        # fixme - remove?
        assert not (np.isnan(orig) | np.isnan(dest)).any()

        # only working with numpy in here
        orig = np.asanyarray(orig).astype(int)
        dest = np.asanyarray(dest).astype(int)

        mapped_orig = self.offset_mapper.map(orig)
        mapped_dest = self.offset_mapper.map(dest)
        result = self.data[mapped_orig, mapped_dest]

        # FIXME - should return nan if not in skim (negative indices wrap around)
        # NOT_IN_SKIM = np.nan
        # in_skim = \
        #     (mapped_orig <0) & (mapped_orig < self.data.shape[0]) & \
        #     (mapped_dest <0) & (mapped_dest < self.data.shape[0])
        # result = np.where(in_skim, result, NOT_IN_SKIM)

        return result


class SkimDict(object):
    """
    A SkimDict object is a wrapper around a dict of multiple skim objects,
    where each object is identified by a key.  It operates like a
    dictionary - i.e. use brackets to add and get skim objects.

    Note that keys are either strings or tuples of two strings (to support stacking of skims.)
    """

    def __init__(self, skim_data, skim_info):

        self.skim_info = skim_info
        self.skim_data = skim_data

        self.offset_mapper = OffsetMapper()
        self.usage = set()

    def touch(self, key):

        self.usage.add(key)

    def get(self, key):
        """
        Get an available wrapped skim object (not the lookup)

        Parameters
        ----------
        key : hashable
             The key (identifier) for this skim object

        Returns
        -------
        skim: Skim
             The skim object
        """

        block, offset = self.skim_info['block_offsets'].get(key)
        block_data = self.skim_data[block]

        self.touch(key)

        data = block_data[:, :, offset]

        return SkimWrapper(data, self.offset_mapper)

    def wrap(self, left_key, right_key):
        """
        return a SkimDictWrapper for self
        """
        return SkimDictWrapper(self, left_key, right_key)


class SkimDictWrapper(object):
    """
    A SkimDictWrapper object is an access wrapper around a SkimDict of multiple skim objects,
    where each object is identified by a key.  It operates like a
    dictionary - i.e. use brackets to add and get skim objects - but also
    has information on how to lookup against the skim objects.
    Specifically, this object has a dataframe, a left_key and right_key.
    It is assumed that left_key and right_key identify columns in df.  The
    parameter df is usually set by the simulation itself as it's a result of
    interacting choosers and alternatives.

    When the user calls skims[key], key is an identifier for which skim
    to use, and the object automatically looks up impedances of that skim
    using the specified left_key column in df as the origin and
    the right_key column in df as the destination.  In this way, the user
    does not do the O-D lookup by hand and only specifies which skim to use
    for this lookup.  This is the only purpose of this object: to
    abstract away the O-D lookup and use skims by specifying which skim
    to use in the expressions.

    Note that keys are either strings or tuples of two strings (to support stacking of skims.)
    """

    def __init__(self, skim_dict, left_key, right_key):
        self.skim_dict = skim_dict
        self.left_key = left_key
        self.right_key = right_key
        self.df = None

    def set_df(self, df):
        """
        Set the dataframe

        Parameters
        ----------
        df : DataFrame
            The dataframe which contains the origin and destination ids

        Returns
        -------
        Nothing
        """
        self.df = df

    def lookup(self, key, reverse=False):
        """
        Generally not called by the user - use __getitem__ instead

        Parameters
        ----------
        key : hashable
             The key (identifier) for this skim object

        od : bool (optional)
            od=True means lookup standard origin-destination skim value
            od=False means lookup destination-origin skim value

        Returns
        -------
        impedances: pd.Series
            A Series of impedances which are elements of the Skim object and
            with the same index as df
        """

        # The skim object to perform the lookup
        # using df[left_key] as the origin and df[right_key] as the destination
        skim = self.skim_dict.get(key)

        # assert self.df is not None, "Call set_df first"
        # origins = self.df[self.left_key].astype('int')
        # destinations = self.df[self.right_key].astype('int')
        # if self.offset:
        #     origins = origins + self.offset
        #     destinations = destinations + self.offset

        assert self.df is not None, "Call set_df first"

        if reverse:
            s = skim.get(self.df[self.right_key], self.df[self.left_key])
        else:
            s = skim.get(self.df[self.left_key], self.df[self.right_key])

        return pd.Series(s, index=self.df.index)

    def reverse(self, key):
        """
        return skim value in reverse (d-o) direction
        """
        return self.lookup(key, reverse=True)

    def max(self, key):
        """
        return max skim value in either o-d or d-o direction
        """

        skim = self.skim_dict.get(key)

        assert self.df is not None, "Call set_df first"

        s = np.maximum(
            skim.get(self.df[self.right_key], self.df[self.left_key]),
            skim.get(self.df[self.left_key], self.df[self.right_key])
        )

        return pd.Series(s, index=self.df.index)

    def __getitem__(self, key):
        """
        Get the (df implicit) lookup for an available skim object

        Parameters
        ----------
        key : hashable
             The key (identifier) for the skim object

        Returns
        -------
        impedances: pd.Series
            A Series of impedances which are elements of the Skim object and
            with the same index as df
        """
        return self.lookup(key)


class SkimStack(object):

    def __init__(self, skim_dict):

        self.offset_mapper = skim_dict.offset_mapper
        self.skim_dict = skim_dict

        # - key1_blocks dict maps key1 to block number
        # DISTWALK: 0,
        # DRV_COM_WLK_BOARDS: 0, ...
        key1_block_offsets = skim_dict.skim_info['key1_block_offsets']
        self.key1_blocks = {k: v[0] for k, v in key1_block_offsets.items()}

        # - skim_dim3 dict maps key1 to dict of key2 absolute offsets into block
        # DRV_COM_WLK_BOARDS: {'MD': 4, 'AM': 3, 'PM': 5}, ...
        block_offsets = skim_dict.skim_info['block_offsets']
        skim_dim3 = OrderedDict()
        for skim_key in block_offsets:

            if not isinstance(skim_key, tuple):
                continue

            key1, key2 = skim_key
            block, offset = block_offsets[skim_key]

            assert block == self.key1_blocks[key1]

            skim_dim3.setdefault(key1, OrderedDict())[key2] = offset

        self.skim_dim3 = skim_dim3

        logger.info("SkimStack.__init__ loaded %s keys with %s total skims"
                    % (len(self.skim_dim3),
                       sum([len(d) for d in self.skim_dim3.values()])))

        self.usage = set()

    def touch(self, key):
        self.usage.add(key)

    def lookup(self, orig, dest, dim3, key):

        orig = self.offset_mapper.map(orig)
        dest = self.offset_mapper.map(dest)

        assert key in self.key1_blocks, "SkimStack key %s missing" % key
        assert key in self.skim_dim3, "SkimStack key %s missing" % key

        block = self.key1_blocks[key]
        stacked_skim_data = self.skim_dict.skim_data[block]
        skim_keys_to_indexes = self.skim_dim3[key]

        self.touch(key)

        # skim_indexes = dim3.map(skim_keys_to_indexes).astype('int')
        # this should be faster than map
        skim_indexes = np.vectorize(skim_keys_to_indexes.get)(dim3)

        return stacked_skim_data[orig, dest, skim_indexes]

    def wrap(self, left_key, right_key, skim_key):
        """
        return a SkimStackWrapper for self
        """
        return SkimStackWrapper(stack=self,
                                left_key=left_key, right_key=right_key, skim_key=skim_key)


class SkimStackWrapper(object):
    """
    A SkimStackWrapper object wraps a SkimStack object to add an additional wrinkle of
    lookup functionality.  Upon init the separate skims objects are
    processed into a 3D matrix so that lookup of the different skims can
    be performed quickly for each row in the dataframe.  In this very
    particular formulation, the keys are assumed to be tuples with two
    elements - the second element of which will be taken from the
    different rows in the dataframe.  The first element can then be
    dereferenced like an array.  This is useful, for instance, to have a
    certain skim vary by time of day - the skims are set with keys of
    ('SOV', 'AM"), ('SOV', 'PM') etc.  The time of day is then taken to
    be different for every row in the tours table, and the 'SOV' portion
    of the key can be used in __getitem__.

    To be more explicit, the input is a dictionary of Skims objects, each of
    which contains a 2D matrix.  These are stacked into a 3D matrix with a
    mapping of keys to indexes which is applied using pandas .map to a third
    column in the object dataframe.  The three columns - left_key and
    right_key from the Skims object and skim_key from this one, are then used to
    dereference the 3D matrix.  The tricky part comes in defining the key which
    matches the 3rd dimension of the matrix, and the key which is passed into
    __getitem__ below (i.e. the one used in the specs).  By convention,
    every key in the Skims object that is passed in MUST be a tuple with 2
    items.  The second item in the tuple maps to the items in the dataframe
    referred to by the skim_key column and the first item in the tuple is
    then available to pass directly to __getitem__.

    The sum conclusion of this is that in the specs, you can say something
    like out_skim['SOV'] and it will automatically dereference the 3D matrix
    using origin, destination, and time of day.

    Parameters
    ----------
    skims: Skims
        This is the Skims object to wrap
    skim_key : str
        This identifies the column in the dataframe which is used to
        select among Skim object using the SECOND item in each tuple (see
        above for a more complete description)
    """

    def __init__(self, stack, left_key, right_key, skim_key):

        self.stack = stack

        self.left_key = left_key
        self.right_key = right_key
        self.skim_key = skim_key
        self.df = None

    def set_df(self, df):
        """
        Set the dataframe

        Parameters
        ----------
        df : DataFrame
            The dataframe which contains the origin and destination ids

        Returns
        -------
        Nothing
        """
        self.df = df

    def __getitem__(self, key):
        """
        Get an available skim object

        Parameters
        ----------
        key : hashable
             The key (identifier) for this skim object

        Returns
        -------
        skim: Skim
             The skim object
        """

        assert self.df is not None, "Call set_df first"
        orig = self.df[self.left_key].astype('int')
        dest = self.df[self.right_key].astype('int')
        dim3 = self.df[self.skim_key]

        skim_values = self.stack.lookup(orig, dest, dim3, key)

        return pd.Series(skim_values, self.df.index)


class DataFrameMatrix(object):
    """
    Utility class to allow a pandas dataframe to be treated like a 2-D array,
    indexed by rowid, colname

    For use in vectorized expressions where the desired values depend on both a row column selector
    e.g. size_terms.get(df.dest_taz, df.purpose)

    ::

      df = pd.DataFrame({'a': [1,2,3,4,5], 'b': [10,20,30,40,50]}, index=[100,101,102,103,104])

      dfm = DataFrameMatrix(df)

      dfm.get(row_ids=[100,100,103], col_ids=['a', 'b', 'a'])

      returns [1, 10,  4]

    """

    def __init__(self, df):
        """

        Parameters
        ----------
        df - pandas dataframe of uniform type
        """

        self.df = df
        self.data = df.values

        self.offset_mapper = OffsetMapper()
        self.offset_mapper.set_offset_list(list(df.index))

        self.cols_to_indexes = {k: v for v, k in enumerate(df.columns)}

    def get(self, row_ids, col_ids):
        """

        Parameters
        ----------
        row_ids - list of row_ids (df index values)
        col_ids - list of column names, one per row_id,
                  specifying column from which the value for that row should be retrieved

        Returns
        -------

        series with one row per row_id, with the value from the column specified in col_ids

        """
        # col_indexes = segments.map(self.cols_to_indexes).astype('int')
        # this should be faster than map
        col_indexes = np.vectorize(self.cols_to_indexes.get)(col_ids)

        row_indexes = self.offset_mapper.map(np.asanyarray(row_ids))

        result = self.data[row_indexes, col_indexes]

        # FIXME - if ids (or col_ids?) is a series, return series with same index?
        if isinstance(row_ids, pd.Series):
            result = pd.Series(result, index=row_ids.index)

        return result



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd
import numpy as np

from activitysim.core import tracing
from activitysim.core import config
from activitysim.core import inject
from activitysim.core import pipeline
from activitysim.core import simulate
from activitysim.core.mem import force_garbage_collect
from activitysim.core.util import assign_in_place

from .util.mode import run_tour_mode_choice_simulate
from .util import estimation

logger = logging.getLogger(__name__)

"""
Tour mode choice is run for all tours to determine the transportation mode that
will be used for the tour
"""


def write_coefficient_template(model_settings):
    coefficients = simulate.read_model_coefficients(model_settings)

    coefficients = coefficients.transpose()
    coefficients.columns.name = None

    template = coefficients.copy()

    coef_names = []
    coef_values = []

    for c in coefficients.columns:

        values = coefficients[c]
        unique_values = values.unique()

        for uv in unique_values:

            if len(unique_values) == 1:
                uv_coef_name = c + '_all'
            else:
                uv_coef_name = c + '_' + '_'.join(values[values == uv].index.values)

            coef_names.append(uv_coef_name)
            coef_values.append(uv)

            template[c] = template[c].where(values != uv, uv_coef_name)

    refactored_coefficients = pd.DataFrame({'coefficient_name': coef_names,  'value': coef_values})
    refactored_coefficients.value = refactored_coefficients.value.astype(np.float32)
    print(refactored_coefficients)

    template = template.transpose()
    template.to_csv(
        config.output_file_path('tour_mode_choice_coefficients_template.csv'),
        mode='w', index=True, header=True)

    refactored_coefficients.to_csv(
        config.output_file_path('tour_mode_choice_refactored_coefficients.csv'),
        mode='w', index=False, header=True)


@inject.step()
def tour_mode_choice_simulate(tours, persons_merged,
                              skim_dict, skim_stack,
                              chunk_size,
                              trace_hh_id):
    """
    Tour mode choice simulate
    """
    trace_label = 'tour_mode_choice'
    model_settings_file_name = 'tour_mode_choice.yaml'
    model_settings = config.read_model_settings(model_settings_file_name)

    logsum_column_name = model_settings.get('MODE_CHOICE_LOGSUM_COLUMN_NAME')
    mode_column_name = 'tour_mode'  # FIXME - should be passed in?

    primary_tours = tours.to_frame()
    assert not (primary_tours.tour_category == 'atwork').any()

    persons_merged = persons_merged.to_frame()

    constants = config.get_model_constants(model_settings)

    logger.info("Running %s with %d tours" % (trace_label, primary_tours.shape[0]))

    tracing.print_summary('tour_types',
                          primary_tours.tour_type, value_counts=True)

    primary_tours_merged = pd.merge(primary_tours, persons_merged, left_on='person_id',
                                    right_index=True, how='left', suffixes=('', '_r'))

    # setup skim keys
    orig_col_name = 'TAZ'
    dest_col_name = 'destination'
    out_time_col_name = 'start'
    in_time_col_name = 'end'
    odt_skim_stack_wrapper = skim_stack.wrap(left_key=orig_col_name, right_key=dest_col_name,
                                             skim_key='out_period')
    dot_skim_stack_wrapper = skim_stack.wrap(left_key=dest_col_name, right_key=orig_col_name,
                                             skim_key='in_period')
    odr_skim_stack_wrapper = skim_stack.wrap(left_key=orig_col_name, right_key=dest_col_name,
                                             skim_key='in_period')
    dor_skim_stack_wrapper = skim_stack.wrap(left_key=dest_col_name, right_key=orig_col_name,
                                             skim_key='out_period')
    od_skim_stack_wrapper = skim_dict.wrap(orig_col_name, dest_col_name)

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "odr_skims": odr_skim_stack_wrapper,
        "dor_skims": dor_skim_stack_wrapper,
        "od_skims": od_skim_stack_wrapper,
        'orig_col_name': orig_col_name,
        'dest_col_name': dest_col_name,
        'out_time_col_name': out_time_col_name,
        'in_time_col_name': in_time_col_name
    }

    estimator = estimation.manager.begin_estimation('tour_mode_choice')
    if estimator:
        estimator.write_coefficients(simulate.read_model_coefficients(model_settings))
        estimator.write_coefficients_template(simulate.read_model_coefficient_template(model_settings))
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        # FIXME run_tour_mode_choice_simulate writes choosers post-annotation

    choices_list = []
    primary_tours_merged['primary_purpose'] = \
        primary_tours_merged.tour_type.where((primary_tours_merged.tour_type != 'school') |
                                             ~primary_tours_merged.is_university, 'univ')

    for primary_purpose, tours_segment in primary_tours_merged.groupby('primary_purpose'):

        logger.info("tour_mode_choice_simulate primary_purpose '%s' (%s tours)" %
                    (primary_purpose, len(tours_segment.index), ))

        # name index so tracing knows how to slice
        assert tours_segment.index.name == 'tour_id'

        choices_df = run_tour_mode_choice_simulate(
            tours_segment,
            primary_purpose, model_settings,
            mode_column_name=mode_column_name,
            logsum_column_name=logsum_column_name,
            skims=skims,
            constants=constants,
            estimator=estimator,
            chunk_size=chunk_size,
            trace_label=tracing.extend_trace_label(trace_label, primary_purpose),
            trace_choice_name='tour_mode_choice')

        tracing.print_summary('tour_mode_choice_simulate %s choices_df' % primary_purpose,
                              choices_df.tour_mode, value_counts=True)

        choices_list.append(choices_df)

        # FIXME - force garbage collection
        force_garbage_collect()

    choices_df = pd.concat(choices_list)

    if estimator:
        estimator.write_choices(choices_df.tour_mode)
        choices_df.tour_mode = estimator.get_survey_values(choices_df.tour_mode, 'tours', 'tour_mode')
        estimator.write_override_choices(choices_df.tour_mode)
        estimator.end_estimation()

    tracing.print_summary('tour_mode_choice_simulate all tour type choices',
                          choices_df.tour_mode, value_counts=True)

    # so we can trace with annotations
    assign_in_place(primary_tours, choices_df)

    # but only keep mode choice col
    all_tours = tours.to_frame()
    assign_in_place(all_tours, choices_df)

    pipeline.replace_table("tours", all_tours)

    if trace_hh_id:
        tracing.trace_df(primary_tours,
                         label=tracing.extend_trace_label(trace_label, mode_column_name),
                         slicer='tour_id',
                         index_label='tour_id',
                         warn_if_empty=True)



# ActivitySim
# See full license in LICENSE.txt.
import logging

from activitysim.core import pipeline
from activitysim.core import inject
from activitysim.core import tracing

from activitysim.core.input import read_input_table

logger = logging.getLogger(__name__)


def read_raw_persons(households):

    df = read_input_table("persons")

    if inject.get_injectable('households_sliced', False):
        # keep all persons in the sampled households
        df = df[df.household_id.isin(households.index)]

    return df


@inject.table()
def persons(households, trace_hh_id):

    df = read_raw_persons(households)

    logger.info("loaded persons %s" % (df.shape,))

    # replace table function with dataframe
    inject.add_table('persons', df)

    pipeline.get_rn_generator().add_channel('persons', df)

    if trace_hh_id:
        tracing.register_traceable_table('persons', df)
        tracing.trace_df(df, "raw.persons", warn_if_empty=True)

    return df


# another common merge for persons
@inject.table()
def persons_merged(persons, households, land_use, accessibility):
    return inject.merge_tables(persons.name, tables=[
        persons, households, land_use, accessibility])



# ActivitySim
# See full license in LICENSE.txt.
import pandas as pd

from activitysim.core import simulate
from activitysim.core import config

from . import expressions
from . import estimation


"""
At this time, these utilities are mostly for transforming the mode choice
spec, which is more complicated than the other specs, into something that
looks like the other specs.
"""


def mode_choice_simulate(
        choosers, spec, nest_spec, skims, locals_d,
        chunk_size,
        mode_column_name,
        logsum_column_name,
        trace_label,
        trace_choice_name,
        estimator=None):

    want_logsums = logsum_column_name is not None

    choices = simulate.simple_simulate(
        choosers=choosers,
        spec=spec,
        nest_spec=nest_spec,
        skims=skims,
        locals_d=locals_d,
        chunk_size=chunk_size,
        want_logsums=want_logsums,
        trace_label=trace_label,
        trace_choice_name=trace_choice_name,
        estimator=estimator)

    # for consistency, always return dataframe, whether or not logsums were requested
    if isinstance(choices, pd.Series):
        choices = choices.to_frame('choice')

    choices.rename(columns={'logsum': logsum_column_name,
                            'choice': mode_column_name},
                   inplace=True)

    alts = spec.columns
    choices[mode_column_name] = \
        choices[mode_column_name].map(dict(list(zip(list(range(len(alts))), alts))))

    return choices


def run_tour_mode_choice_simulate(
        choosers,
        tour_purpose, model_settings,
        mode_column_name,
        logsum_column_name,
        skims,
        constants,
        estimator,
        chunk_size,
        trace_label=None, trace_choice_name=None):
    """
    This is a utility to run a mode choice model for each segment (usually
    segments are tour/trip purposes).  Pass in the tours/trip that need a mode,
    the Skim object, the spec to evaluate with, and any additional expressions
    you want to use in the evaluation of variables.
    """

    spec = simulate.read_model_spec(file_name=model_settings['SPEC'])
    coefficients = simulate.get_segment_coefficients(model_settings, tour_purpose)
    spec = simulate.eval_coefficients(spec, coefficients, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)
    nest_spec = simulate.eval_nest_coefficients(nest_spec, coefficients)

    locals_dict = {}
    locals_dict.update(constants)
    locals_dict.update(skims)

    # constrained coefficients can appear in expressions
    locals_dict.update(coefficients)

    assert ('in_period' not in choosers) and ('out_period' not in choosers)
    in_time = skims['in_time_col_name']
    out_time = skims['out_time_col_name']
    choosers['in_period'] = expressions.skim_time_period_label(choosers[in_time])
    choosers['out_period'] = expressions.skim_time_period_label(choosers[out_time])

    expressions.annotate_preprocessors(
        choosers, locals_dict, skims,
        model_settings, trace_label)

    if estimator:
        # write choosers after annotation
        estimator.write_choosers(choosers)

    choices = mode_choice_simulate(
        choosers=choosers,
        spec=spec,
        nest_spec=nest_spec,
        skims=skims,
        locals_d=locals_dict,
        chunk_size=chunk_size,
        mode_column_name=mode_column_name,
        logsum_column_name=logsum_column_name,
        trace_label=trace_label,
        trace_choice_name=trace_choice_name,
        estimator=estimator)

    return choices



# ActivitySim
# See full license in LICENSE.txt.

import numpy as np
import pandas as pd
import numpy.testing as npt
import pandas.testing as pdt
import pytest

from .. import skim


@pytest.fixture
def data():
    return np.arange(100, dtype='int').reshape((10, 10))


def test_basic(data):
    sk = skim.SkimWrapper(data)

    orig = [5, 9, 1]
    dest = [2, 9, 6]

    npt.assert_array_equal(
        sk.get(orig, dest),
        [52, 99, 16])


def test_offset_int(data):
    sk = skim.SkimWrapper(data, skim.OffsetMapper(-1))

    orig = [6, 10, 2]
    dest = [3, 10, 7]

    npt.assert_array_equal(
        sk.get(orig, dest),
        [52, 99, 16])


def test_offset_list(data):

    offset_mapper = skim.OffsetMapper()
    offset_mapper.set_offset_list([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    # should have figured out it could use an int offset instead of list
    assert offset_mapper.offset_int == -1

    offset_mapper = skim.OffsetMapper()
    offset_mapper.set_offset_list([10, 20, 30, 40, 50, 60, 70, 80, 90, 100])

    sk = skim.SkimWrapper(data, offset_mapper)

    orig = [60, 100, 20]
    dest = [30, 100, 70]

    npt.assert_array_equal(
        sk.get(orig, dest),
        [52, 99, 16])


# fixme - nan support disabled in skim.py (not sure we need it?)
# def test_skim_nans(data):
#     sk = skim.SkimWrapper(data)
#
#     orig = [5, np.nan, 1, 2]
#     dest = [np.nan, 9, 6, 4]
#
#     npt.assert_array_equal(
#         sk.get(orig, dest),
#         [np.nan, np.nan, 16, 24])


def test_skims(data):

    skims_shape = data.shape + (2,)

    skim_data = np.zeros(skims_shape, dtype=data.dtype)
    skim_data[:, :, 0] = data
    skim_data[:, :, 1] = data*10

    skim_info = {
        'block_offsets': {'AM': (0, 0), 'PM': (0, 1)}
    }

    skim_dict = skim.SkimDict([skim_data], skim_info)

    skims = skim_dict.wrap("taz_l", "taz_r")

    df = pd.DataFrame({
        "taz_l": [1, 9, 4],
        "taz_r": [2, 3, 7],
    })

    skims.set_df(df)

    pdt.assert_series_equal(
        skims["AM"],
        pd.Series(
            [12, 93, 47],
            index=[0, 1, 2]
        ).astype(data.dtype)
    )

    pdt.assert_series_equal(
        skims["PM"],
        pd.Series(
            [120, 930, 470],
            index=[0, 1, 2]
        ).astype(data.dtype)
    )


def test_3dskims(data):

    skims_shape = data.shape + (2,)

    skim_data = np.zeros(skims_shape, dtype=int)
    skim_data[:, :, 0] = data
    skim_data[:, :, 1] = data*10

    skim_info = {
        'block_offsets': {('SOV', 'AM'): (0, 0), ('SOV', 'PM'): (0, 1)},
        'key1_block_offsets': {'SOV': (0, 0)}
    }
    skim_dict = skim.SkimDict([skim_data], skim_info)

    stack = skim.SkimStack(skim_dict)

    skims3d = stack.wrap(left_key="taz_l", right_key="taz_r", skim_key="period")

    df = pd.DataFrame({
        "taz_l": [1, 9, 4],
        "taz_r": [2, 3, 7],
        "period": ["AM", "PM", "AM"]
    })

    skims3d.set_df(df)

    pdt.assert_series_equal(
        skims3d["SOV"],
        pd.Series(
            [12, 930, 47],
            index=[0, 1, 2]
        ),
        check_dtype=False
    )



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import tracing
from activitysim.core import config
from activitysim.core import inject
from activitysim.core import pipeline
from activitysim.core import simulate

from activitysim.core.mem import force_garbage_collect

from activitysim.core.util import assign_in_place

from .util.mode import run_tour_mode_choice_simulate
from .util import estimation


logger = logging.getLogger(__name__)


@inject.step()
def atwork_subtour_mode_choice(
        tours,
        persons_merged,
        skim_dict, skim_stack,
        chunk_size,
        trace_hh_id):
    """
    At-work subtour mode choice simulate
    """

    trace_label = 'atwork_subtour_mode_choice'

    model_settings_file_name = 'tour_mode_choice.yaml'
    model_settings = config.read_model_settings(model_settings_file_name)

    logsum_column_name = model_settings.get('MODE_CHOICE_LOGSUM_COLUMN_NAME')
    mode_column_name = 'tour_mode'

    tours = tours.to_frame()
    subtours = tours[tours.tour_category == 'atwork']

    # - if no atwork subtours
    if subtours.shape[0] == 0:
        tracing.no_results(trace_label)
        return

    subtours_merged = \
        pd.merge(subtours, persons_merged.to_frame(),
                 left_on='person_id', right_index=True, how='left')

    constants = config.get_model_constants(model_settings)

    logger.info("Running %s with %d subtours" % (trace_label, subtours_merged.shape[0]))

    tracing.print_summary('%s tour_type' % trace_label,
                          subtours_merged.tour_type, value_counts=True)

    # setup skim keys
    orig_col_name = 'workplace_taz'
    dest_col_name = 'destination'
    out_time_col_name = 'start'
    in_time_col_name = 'end'
    odt_skim_stack_wrapper = skim_stack.wrap(left_key=orig_col_name, right_key=dest_col_name,
                                             skim_key='out_period')
    dot_skim_stack_wrapper = skim_stack.wrap(left_key=dest_col_name, right_key=orig_col_name,
                                             skim_key='in_period')
    odr_skim_stack_wrapper = skim_stack.wrap(left_key=orig_col_name, right_key=dest_col_name,
                                             skim_key='in_period')
    dor_skim_stack_wrapper = skim_stack.wrap(left_key=dest_col_name, right_key=orig_col_name,
                                             skim_key='out_period')
    od_skim_stack_wrapper = skim_dict.wrap(orig_col_name, dest_col_name)

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "odr_skims": odr_skim_stack_wrapper,
        "dor_skims": dor_skim_stack_wrapper,
        "od_skims": od_skim_stack_wrapper,
        'orig_col_name': orig_col_name,
        'dest_col_name': dest_col_name,
        'out_time_col_name': out_time_col_name,
        'in_time_col_name': in_time_col_name
    }

    estimator = estimation.manager.begin_estimation('atwork_subtour_mode_choice')
    if estimator:
        estimator.write_coefficients(simulate.read_model_coefficients(model_settings))
        estimator.write_coefficients_template(simulate.read_model_coefficient_template(model_settings))
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        # FIXME run_tour_mode_choice_simulate writes choosers post-annotation

    choices_df = run_tour_mode_choice_simulate(
        subtours_merged,
        tour_purpose='atwork', model_settings=model_settings,
        mode_column_name=mode_column_name,
        logsum_column_name=logsum_column_name,
        skims=skims,
        constants=constants,
        estimator=estimator,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name='tour_mode_choice')

    if estimator:
        estimator.write_choices(choices_df[mode_column_name])
        choices_df[mode_column_name] = \
            estimator.get_survey_values(choices_df[mode_column_name], 'tours', mode_column_name)
        estimator.write_override_choices(choices_df[mode_column_name])
        estimator.end_estimation()

    tracing.print_summary('%s choices' % trace_label, choices_df[mode_column_name], value_counts=True)

    assign_in_place(tours, choices_df)
    pipeline.replace_table("tours", tours)

    if trace_hh_id:
        tracing.trace_df(tours[tours.tour_category == 'atwork'],
                         label=tracing.extend_trace_label(trace_label, mode_column_name),
                         slicer='tour_id',
                         index_label='tour_id')

    force_garbage_collect()



# ActivitySim
# See full license in LICENSE.txt.

import os.path

import pandas as pd
import pandas.testing as pdt
import pytest

from .. import cdap

from activitysim.core import simulate


@pytest.fixture(scope='module')
def data_dir():
    return os.path.join(os.path.dirname(__file__), 'data')


@pytest.fixture(scope='module')
def configs_dir():
    return os.path.join(os.path.dirname(__file__), 'configs')


@pytest.fixture(scope='module')
def people(data_dir):
    return pd.read_csv(
        os.path.join(data_dir, 'people.csv'),
        index_col='id')


@pytest.fixture(scope='module')
def cdap_indiv_and_hhsize1(configs_dir):
    return simulate.read_model_spec(file_name='cdap_indiv_and_hhsize1.csv', spec_dir=configs_dir)


@pytest.fixture(scope='module')
def cdap_interaction_coefficients(configs_dir):
    f = os.path.join(configs_dir, 'cdap_interaction_coefficients.csv')
    coefficients = pd.read_csv(f, comment='#')
    coefficients = cdap.preprocess_interaction_coefficients(coefficients)
    return coefficients


@pytest.fixture(scope='module')
def individual_utils(
        people, cdap_indiv_and_hhsize1):
    return cdap.individual_utilities(people, cdap_indiv_and_hhsize1, locals_d=None)


def test_bad_coefficients(configs_dir):

    f = os.path.join(configs_dir, 'cdap_interaction_coefficients.csv')
    coefficients = pd.read_csv(f, comment='#')

    coefficients.loc[2, 'activity'] = 'AA'

    with pytest.raises(RuntimeError) as excinfo:
        coefficients = cdap.preprocess_interaction_coefficients(coefficients)
    assert "Expect only M, N, or H" in str(excinfo.value)


def test_assign_cdap_rank(people):

    cdap.assign_cdap_rank(people)

    expected = pd.Series(
        [1, 1, 1, 2, 2, 1, 3, 1, 2, 1, 3, 2, 1, 3, 2, 4, 1, 3, 4, 2],
        index=people.index
    )

    pdt.assert_series_equal(people['cdap_rank'], expected, check_dtype=False, check_names=False)


def test_individual_utilities(people, cdap_indiv_and_hhsize1):

    cdap.assign_cdap_rank(people)
    individual_utils = cdap.individual_utilities(people, cdap_indiv_and_hhsize1, locals_d=None)
    individual_utils = individual_utils[['M', 'N', 'H']]

    expected = pd.DataFrame([
        [2, 0, 0],  # person 1
        [0, 0, 1],  # person 2
        [3, 0, 0],  # person 3
        [3, 0, 0],  # person 4
        [0, 1, 0],  # person 5
        [1, 0, 0],  # person 6
        [1, 0, 0],  # person 7
        [0, 2, 0],  # person 8
        [0, 0, 1],  # person 9
        [2, 0, 0],  # person 10
        [0, 0, 3],  # person 11
        [0, 0, 2],  # person 12
        [3, 0, 0],  # person 13
        [1, 0, 0],  # person 14
        [0, 4, 0],  # person 15
        [0, 4, 0],  # person 16
        [0, 0, 4],  # person 17
        [0, 0, 5],  # person 18
        [50, 0, 4],  # person 19
        [2, 0, 0]  # person 20
        ], index=people.index, columns=cdap_indiv_and_hhsize1.columns)

    pdt.assert_frame_equal(
        individual_utils, expected, check_dtype=False, check_names=False)


def test_build_cdap_spec_hhsize2(people, cdap_indiv_and_hhsize1, cdap_interaction_coefficients):

    hhsize = 2

    cdap.assign_cdap_rank(people)
    indiv_utils = cdap.individual_utilities(people, cdap_indiv_and_hhsize1, locals_d=None)

    choosers = cdap.hh_choosers(indiv_utils, hhsize=hhsize)

    spec = cdap.build_cdap_spec(cdap_interaction_coefficients, hhsize=hhsize, cache=False)

    vars = simulate.eval_variables(spec.index, choosers)

    utils = simulate.compute_utilities(vars, spec)

    expected = pd.DataFrame([
        [0, 3, 0, 3, 7, 3, 0, 3, 0],  # household 3
        [0, 0, 1, 1, 1, 2, 0, 0, 2],  # household 4
        ],
        index=[3, 4],
        columns=['HH', 'HM', 'HN', 'MH', 'MM', 'MN', 'NH', 'NM', 'NN']).astype('float')

    pdt.assert_frame_equal(utils, expected, check_names=False)



# ActivitySim
# See full license in LICENSE.txt.
HHT_NONE = 0
HHT_FAMILY_MARRIED = 1
HHT_FAMILY_MALE = 2
HHT_FAMILY_FEMALE = 3
HHT_NONFAMILY_MALE_ALONE = 4
HHT_NONFAMILY_MALE_NOTALONE = 5
HHT_NONFAMILY_FEMALE_ALONE = 6
HHT_NONFAMILY_FEMALE_NOTALONE = 7

# convenience for expression files
HHT_NONFAMILY = [4, 5, 6, 7]
HHT_FAMILY = [1, 2, 3]

PSTUDENT_GRADE_OR_HIGH = 1
PSTUDENT_UNIVERSITY = 2
PSTUDENT_NOT = 3

GRADE_SCHOOL_MAX_AGE = 14
GRADE_SCHOOL_MIN_AGE = 5

SCHOOL_SEGMENT_NONE = 0
SCHOOL_SEGMENT_GRADE = 1
SCHOOL_SEGMENT_HIGH = 2
SCHOOL_SEGMENT_UNIV = 3

INCOME_SEGMENT_LOW = 1
INCOME_SEGMENT_MED = 2
INCOME_SEGMENT_HIGH = 3
INCOME_SEGMENT_VERYHIGH = 4

PEMPLOY_FULL = 1
PEMPLOY_PART = 2
PEMPLOY_NOT = 3
PEMPLOY_CHILD = 4

PTYPE_FULL = 1
PTYPE_PART = 2
PTYPE_UNIVERSITY = 3
PTYPE_NONWORK = 4
PTYPE_RETIRED = 5
PTYPE_DRIVING = 6
PTYPE_SCHOOL = 7
PTYPE_PRESCHOOL = 8


# these appear as column headers in non_mandatory_tour_frequency.csv
PTYPE_NAME = {
    PTYPE_FULL: 'PTYPE_FULL',
    PTYPE_PART: 'PTYPE_PART',
    PTYPE_UNIVERSITY: 'PTYPE_UNIVERSITY',
    PTYPE_NONWORK: 'PTYPE_NONWORK',
    PTYPE_RETIRED: 'PTYPE_RETIRED',
    PTYPE_DRIVING: 'PTYPE_DRIVING',
    PTYPE_SCHOOL: 'PTYPE_SCHOOL',
    PTYPE_PRESCHOOL: 'PTYPE_PRESCHOOL'
}

CDAP_ACTIVITY_MANDATORY = 'M'
CDAP_ACTIVITY_NONMANDATORY = 'N'
CDAP_ACTIVITY_HOME = 'H'



# ActivitySim
# See full license in LICENSE.txt.
import logging

from activitysim.core.util import assign_in_place


logger = logging.getLogger(__name__)


def failed_trip_cohorts(trips, failed):

    # outbound trips in a tour with a failed outbound trip
    bad_outbound_trips = \
        trips.outbound & (trips.tour_id.isin(trips.tour_id[failed & trips.outbound]))

    # inbound trips in a tour with a failed inbound trip
    bad_inbound_trips = \
        ~trips.outbound & (trips.tour_id.isin(trips.tour_id[failed & ~trips.outbound]))

    bad_trips = bad_outbound_trips | bad_inbound_trips

    return bad_trips


def flag_failed_trip_leg_mates(trips_df, col_name):
    """
    set boolean flag column of specified name to identify failed trip leg_mates in place
    """

    failed_trip_leg_mates = failed_trip_cohorts(trips_df, trips_df.failed) & ~trips_df.failed
    trips_df.loc[failed_trip_leg_mates, col_name] = True

    # handle outbound and inbound legs independently
    # for ob in [True, False]:
    #     same_leg = (trips_df.outbound == ob)
    #     # tour_ids of all tours with a failed trip in this (outbound or inbound) leg direction
    #     bad_tours = trips_df.tour_id[trips_df.failed & same_leg].unique()
    #     # not-failed leg_mates of all failed trips in this (outbound or inbound) leg direction
    #     failed_trip_leg_mates = same_leg & (trips_df.tour_id.isin(bad_tours)) & ~trips_df.failed
    #     # set the flag column
    #     trips_df.loc[failed_trip_leg_mates, col_name] = True


def cleanup_failed_trips(trips):
    """
    drop failed trips and cleanup fields in leg_mates:

    trip_num        assign new ordinal trip num after failed trips are dropped
    trip_count      assign new count of trips in leg, sans failed trips
    first           update first flag as we may have dropped first trip (last trip can't fail)
    next_trip_id    assign id of next trip in leg after failed trips are dropped
    """

    if trips.failed.any():
        logger.warning("cleanup_failed_trips dropping %s failed trips" % trips.failed.sum())

        trips['patch'] = False
        flag_failed_trip_leg_mates(trips, 'patch')

        # drop the original failures
        trips = trips[~trips.failed]

        # increasing trip_id order
        patch_trips = trips[trips.patch].sort_index()

        # recompute fields dependent on trip_num sequence
        grouped = patch_trips.groupby(['tour_id', 'outbound'])
        patch_trips['trip_num'] = grouped.cumcount() + 1
        patch_trips['trip_count'] = patch_trips['trip_num'] + grouped.cumcount(ascending=False)

        assign_in_place(trips, patch_trips[['trip_num', 'trip_count']])

        del trips['patch']

    del trips['failed']

    return trips



import os

import pandas as pd
import pytest
import yaml

from activitysim.core import inject
from activitysim.abm.models.util import expressions


@pytest.fixture(scope="session")
def config_path():
    return os.path.join(os.path.dirname(__file__), 'configs_test_misc')


def test_30_minute_windows(config_path):
    with open(os.path.join(config_path, 'settings_30_min.yaml')) as f:
        settings = yaml.load(f, Loader=yaml.SafeLoader)

    inject.add_injectable("settings", settings)

    assert expressions.skim_time_period_label(1) == 'EA'
    assert expressions.skim_time_period_label(16) == 'AM'
    assert expressions.skim_time_period_label(24) == 'MD'
    assert expressions.skim_time_period_label(36) == 'PM'
    assert expressions.skim_time_period_label(46) == 'EV'

    pd.testing.assert_series_equal(
        expressions.skim_time_period_label(pd.Series([1, 16, 24, 36, 46])),
        pd.Series(['EA', 'AM', 'MD', 'PM', 'EV']))


def test_60_minute_windows(config_path):
    with open(os.path.join(config_path, 'settings_60_min.yaml')) as f:
        settings = yaml.load(f, Loader=yaml.SafeLoader)

    inject.add_injectable("settings", settings)

    assert expressions.skim_time_period_label(1) == 'EA'
    assert expressions.skim_time_period_label(8) == 'AM'
    assert expressions.skim_time_period_label(12) == 'MD'
    assert expressions.skim_time_period_label(18) == 'PM'
    assert expressions.skim_time_period_label(23) == 'EV'

    pd.testing.assert_series_equal(
        expressions.skim_time_period_label(pd.Series([1, 8, 12, 18, 23])),
        pd.Series(['EA', 'AM', 'MD', 'PM', 'EV']))


def test_1_week_time_window():
    settings = {
        'skim_time_periods': {
            'time_window': 10080,  # One Week
            'period_minutes': 1440,  # One Day
            'periods': [
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7
            ],
            'labels': ['Sunday', 'Monday', 'Tuesday', 'Wednesday',
                       'Thursday', 'Friday', 'Saturday']
        }
    }

    inject.add_injectable("settings", settings)

    assert expressions.skim_time_period_label(1) == 'Sunday'
    assert expressions.skim_time_period_label(2) == 'Monday'
    assert expressions.skim_time_period_label(3) == 'Tuesday'
    assert expressions.skim_time_period_label(4) == 'Wednesday'
    assert expressions.skim_time_period_label(5) == 'Thursday'
    assert expressions.skim_time_period_label(6) == 'Friday'
    assert expressions.skim_time_period_label(7) == 'Saturday'

    weekly_series = expressions.skim_time_period_label(pd.Series([1, 2, 3, 4, 5, 6, 7]))

    pd.testing.assert_series_equal(weekly_series,
                                   pd.Series(['Sunday', 'Monday', 'Tuesday', 'Wednesday',
                                              'Thursday', 'Friday', 'Saturday']))


def test_future_warning(config_path):
    with open(os.path.join(config_path, 'settings_60_min.yaml')) as f:
        settings = yaml.load(f, Loader=yaml.SafeLoader)

    settings['skim_time_periods']['hours'] = settings['skim_time_periods'].pop('periods')

    inject.add_injectable("settings", settings)

    with pytest.warns(FutureWarning) as warning_test:
        expressions.skim_time_period_label(1)
